# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest

from hyputils.memex._compat import text_type
from hyputils.memex.util import uri

TEST_URLS = [
    # Should replace http and https protocol with httpx
    ("https://example.org", "httpx://example.org"),
    ("http://example.org", "httpx://example.org"),
    # but not when scheme is not http or https
    ("ftp://example.org", "ftp://example.org"),
    ("file://example.org", "file://example.org"),
    # Should strip https://via.hypothes.is/ from the start of URIs
    ("https://via.hypothes.is/https://example.com", "httpx://example.com"),
    ("https://via.hypothes.is/http://foo.com/bar/", "httpx://foo.com/bar"),
    # but not when the URI isn't a proxied one
    ("https://via.hypothes.is", "httpx://via.hypothes.is"),
    ("https://via.hypothes.is/sample", "httpx://via.hypothes.is/sample"),
    # Should leave URNs as they are
    ("urn:doi:10.0001/12345", "urn:doi:10.0001/12345"),
    # Should leave http(s) URLs with no hostname as they are
    ("http:///path/to/page", "http:///path/to/page"),
    ("https:///path/to/page", "https:///path/to/page"),
    # Should treat messed up urlencoded strings as opaque and return them as
    # is. This is not valid urlencoding and trying to deal with it is going to
    # cause more problems than solutions (for example: is "%2F" in the path
    # section a path segment delimiter or a literal solidus?).
    ("http%3A%2F%2Fexample.com", "http%3A%2F%2Fexample.com"),
    ("http%3A%2F%2Fexample.com%2Ffoo%2F", "http%3A%2F%2Fexample.com%2Ffoo%2F"),
    # Should leave already-normalized URLs as they are
    ("http://example.com", "httpx://example.com"),
    ("https://foo.bar.org", "httpx://foo.bar.org"),
    # Should case-normalize scheme
    ("HtTp://example.com", "httpx://example.com"),
    ("HTTP://example.com", "httpx://example.com"),
    ("HtTpS://example.com", "httpx://example.com"),
    ("HTTPS://example.com", "httpx://example.com"),
    # Should case-normalize hostname
    ("http://EXAMPLE.COM", "httpx://example.com"),
    ("http://EXampLE.COM", "httpx://example.com"),
    # Should leave userinfo case alone
    ("http://Alice:p4SSword@example.com", "httpx://Alice:p4SSword@example.com"),
    ("http://BOB@example.com", "httpx://BOB@example.com"),
    # Should leave path case alone
    ("http://example.com/FooBar", "httpx://example.com/FooBar"),
    ("http://example.com/FOOBAR", "httpx://example.com/FOOBAR"),
    # Should strip URL fragments
    ("http://example.com#", "httpx://example.com"),
    ("http://example.com#bar", "httpx://example.com"),
    ("http://example.com/path#", "httpx://example.com/path"),
    ("http://example.com/path#!/hello/world", "httpx://example.com/path"),
    # Should remove default ports
    ("http://example.com:80", "httpx://example.com"),
    ("http://example.com:81", "httpx://example.com:81"),
    ("http://example.com:443", "httpx://example.com:443"),
    ("https://example.com:443", "httpx://example.com"),
    ("https://example.com:1443", "httpx://example.com:1443"),
    ("https://example.com:80", "httpx://example.com:80"),
    ("http://[fe80::3e15:c2ff:fed6:d198]:80", "httpx://[fe80::3e15:c2ff:fed6:d198]"),
    ("https://[fe80::3e15:c2ff:fed6:d198]:443", "httpx://[fe80::3e15:c2ff:fed6:d198]"),
    # Path: remove trailing slashes
    ("http://example.com/", "httpx://example.com"),
    ("http://example.com/////", "httpx://example.com"),
    ("http://example.com/foo/bar/baz/", "httpx://example.com/foo/bar/baz"),
    ("http://example.com/foo/bar/baz/////", "httpx://example.com/foo/bar/baz"),
    # Path: ensure UNRESERVED characters are decoded
    ("http://example.com/%7Ealice", "httpx://example.com/~alice"),
    ("http://example.com/~alice", "httpx://example.com/~alice"),
    (
        "http://example.com/goes%2Dto%2Dwonderland",
        "httpx://example.com/goes-to-wonderland",
    ),
    ("http://example.com/goes-to-wonderland", "httpx://example.com/goes-to-wonderland"),
    ("http://example.com/%41%42%43/%31%32%33", "httpx://example.com/ABC/123"),
    ("http://example.com/hello%2Bworld", "httpx://example.com/hello+world"),
    ("http://example.com/hello+world", "httpx://example.com/hello+world"),
    (
        "http://example.com/%3A%40%2D%2E%5F%7E%21%24%26%27%28%29%2A%2B%2C%3D%3B",
        "httpx://example.com/:@-._~!$&'()*+,=;",
    ),
    ("http://example.com/:@-._~!$&'()*+,=;", "httpx://example.com/:@-._~!$&'()*+,=;"),
    # Path: ensure RESERVED characters are encoded
    ("http://example.com/foo%2Fbar", "httpx://example.com/foo%2Fbar"),
    ("http://example.com/foo%3Fbar", "httpx://example.com/foo%3Fbar"),
    ("http://example.com/foo%5Bbar%5D", "httpx://example.com/foo%5Bbar%5D"),
    ("http://example.com/foo[bar]", "httpx://example.com/foo%5Bbar%5D"),
    # Path: ensure OTHER characters are encoded
    ("http://example.com/كذا", "httpx://example.com/%D9%83%D8%B0%D8%A7"),
    ("http://example.com/snowman/☃", "httpx://example.com/snowman/%E2%98%83"),
    # Path: normalize case of encodings
    ("http://example.com/case%2fnormalized", "httpx://example.com/case%2Fnormalized"),
    # Query: remove empty
    ("http://example.com?", "httpx://example.com"),
    # Query: Preserve the query string if we can't parse it
    ("http://example.com?&", "httpx://example.com?&"),
    ("http://example.com?foo=&", "httpx://example.com?foo=&"),
    # Query: ensure UNRESERVED characters are decoded
    ("http://example.com?foo%7Ebar=baz", "httpx://example.com?foo~bar=baz"),
    ("http://example.com?foo~bar=baz", "httpx://example.com?foo~bar=baz"),
    ("http://example.com?foo=bar%7Ebaz", "httpx://example.com?foo=bar~baz"),
    ("http://example.com?foo=bar~baz", "httpx://example.com?foo=bar~baz"),
    (
        "http://example.com?-._~:@!$'()*,=-._~:@!$='()*,",
        "httpx://example.com?-._~:@!$'()*,=-._~:@!$='()*,",
    ),
    (
        "http://example.com?%2D%2E%5F%7E%3A%40%21%24%27%28%29%2A%2C=%2D%2E%5F%7E%3A%40%21%24%3D%27%28%29%2A%2C",
        "httpx://example.com?-._~:@!$'()*,=-._~:@!$='()*,",
    ),
    # Query: ensure RESERVED characters are encoded
    ("http://example.com?foo bar=baz", "httpx://example.com?foo+bar=baz"),
    ("http://example.com?foo+bar=baz", "httpx://example.com?foo+bar=baz"),
    ("http://example.com?foo%20bar=baz", "httpx://example.com?foo+bar=baz"),
    ("http://example.com?foo=bar baz", "httpx://example.com?foo=bar+baz"),
    ("http://example.com?foo=bar+baz", "httpx://example.com?foo=bar+baz"),
    ("http://example.com?foo=bar%20baz", "httpx://example.com?foo=bar+baz"),
    ("http://example.com?foo%5Bbar%5D=baz", "httpx://example.com?foo%5Bbar%5D=baz"),
    ("http://example.com?foo[bar]=baz", "httpx://example.com?foo%5Bbar%5D=baz"),
    ("http://example.com?foo=%5Bbar%5Dbaz", "httpx://example.com?foo=%5Bbar%5Dbaz"),
    ("http://example.com?foo=[bar]baz", "httpx://example.com?foo=%5Bbar%5Dbaz"),
    # Query: ensure OTHER characters are encoded
    (
        "http://example.com?你好世界=γειά σου κόσμος",
        "httpx://example.com?%E4%BD%A0%E5%A5%BD%E4%B8%96%E7%95%8C=%CE%B3%CE%B5%CE%B9%CE%AC+%CF%83%CE%BF%CF%85+%CE%BA%CF%8C%CF%83%CE%BC%CE%BF%CF%82",
    ),
    ("http://example.com?love=♥", "httpx://example.com?love=%E2%99%A5"),
    # Query: normalize case of encodings
    ("http://example.com?love=%e2%99%a5", "httpx://example.com?love=%E2%99%A5"),
    # Query: lexically sort parameters by name
    ("http://example.com?a=1&b=2", "httpx://example.com?a=1&b=2"),
    ("http://example.com?b=2&a=1", "httpx://example.com?a=1&b=2"),
    # Query: preserve relative ordering of multiple params with the same name
    ("http://example.com?b=2&b=3&b=1&a=1", "httpx://example.com?a=1&b=2&b=3&b=1"),
    ("http://example.com?b=&b=3&b=1&a=1", "httpx://example.com?a=1&b=&b=3&b=1"),
    # Query: remove parameters known to be irrelevant for document identity
    ("http://example.com?utm_source=abcde", "httpx://example.com"),
    ("http://example.com?utm_medium=abcde", "httpx://example.com"),
    ("http://example.com?utm_term=abcde", "httpx://example.com"),
    ("http://example.com?utm_content=abcde", "httpx://example.com"),
    ("http://example.com?utm_campaign=abcde", "httpx://example.com"),
    ("http://example.com?utm_source=abcde&utm_medium=wibble", "httpx://example.com"),
    ("http://example.com?a=1&utm_term=foo", "httpx://example.com?a=1"),
    ("http://example.com?a=1&utm_term=foo&b=2", "httpx://example.com?a=1&b=2"),
    ("http://example.com?WT.mc_id=TWT_NatureNews", "httpx://example.com"),
    ("http://example.com?WT.foo=bar", "httpx://example.com"),
    # but don't be over-eager and remove close matches
    ("http://example.com?gclid_foo=abcde", "httpx://example.com?gclid_foo=abcde"),
    ("http://example.com?bar_gclid=abcde", "httpx://example.com?bar_gclid=abcde"),
    ("http://example.com?WT=abcde", "httpx://example.com?WT=abcde"),
]


@pytest.mark.parametrize("url_in,url_out", TEST_URLS)
def test_normalize(url_in, url_out):
    assert uri.normalize(url_in) == url_out


@pytest.mark.parametrize("url,_", TEST_URLS)
def test_normalize_returns_unicode(url, _):
    assert isinstance(uri.normalize(url), text_type)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import datetime
import re

import sqlalchemy as sa
from sqlalchemy.ext.hybrid import Comparator, hybrid_property
from sqlalchemy.ext.declarative import declared_attr

from hyputils.memex._compat import string_types
from hyputils.memex.db import Base
from hyputils.memex.util.user import split_user
from hyputils.memex.security import security

USERNAME_MIN_LENGTH = 3
USERNAME_MAX_LENGTH = 30
USERNAME_PATTERN = "(?i)^[A-Z0-9._]+$"
EMAIL_MAX_LENGTH = 100
DISPLAY_NAME_MAX_LENGTH = 30


def _normalise_username(username):
    # We normalize usernames by dots and case in order to discourage attempts
    # at impersonation.
    return sa.func.lower(sa.func.replace(username, sa.text("'.'"), sa.text("''")))


class UsernameComparator(Comparator):
    """
    Custom comparator for :py:attr:`~h.models.user.User.username`.

    This ensures that all lookups against the username property, such as

        session.query(User).filter_by(username='juanwood')

    use the normalised username for the lookup and appropriately normalise the
    RHS of the query. This means that a query like the one above will
    correctly find a user with a username of "Juan.Wood", for example.
    """

    def operate(self, op, other, **kwargs):
        return op(
            _normalise_username(self.__clause_element__()),
            _normalise_username(other),
            **kwargs
        )


class UserIDComparator(Comparator):
    """
    Custom comparator for :py:attr:`~h.models.user.User.userid`.

    A user's userid is a compound property which depends on their username
    and their authority. A naive comparator for this property would generate
    SQL like the following:

        ... WHERE 'acct:' || username || '@' || authority = ...

    This would be slow, due to the lack of an index on the LHS expression.
    While we could add a functional index on this expression, we can also take
    advantage of the existing index on (normalised_username, authority), which
    is what this comparator does.

    A query such as

        session.query(User).filter_by(userid='acct:luis.silva@example.com')

    will instead generate

        WHERE
            (lower(replace(username,     '.', '')), authority    ) =
            (lower(replace('luis.silva', '.', '')), 'example.com')
    """

    def __init__(self, username, authority):
        self.username = username
        self.authority = authority

    def __clause_element__(self):
        return sa.tuple_(_normalise_username(self.username), self.authority)

    def __eq__(self, other):
        """
        Compare the userid for equality with `other`.

        `other` can be anything plausibly on the RHS of a comparison, which
        can include other SQL clause elements or expressions, as in

            User.userid == sa.tuple_(User.username, Group.authority)

        or literals, as in

            User.userid == 'acct:miruna@example.com'

        We treat the literal case specially, and split the string into
        username and authority ourselves. If the string is not a well-formed
        userid, the comparison will always return False.
        """
        if isinstance(other, string_types):
            try:
                val = split_user(other)
            except ValueError:
                # The value being compared isn't a valid userid
                return False
            else:
                other = sa.tuple_(_normalise_username(val["username"]), val["domain"])
        return self.__clause_element__() == other

    def in_(self, userids):
        others = []
        for userid in userids:
            try:
                val = split_user(userid)
            except ValueError:
                continue

            other = sa.tuple_(_normalise_username(val["username"]), val["domain"])
            others.append(other)

        if not others:
            return False

        return self.__clause_element__().in_(others)


class User(Base):
    __tablename__ = "user"

    @declared_attr
    def __table_args__(cls):  # noqa: N805
        return (
            # (email, authority) must be unique
            sa.UniqueConstraint("email", "authority"),
            # (normalised username, authority) must be unique. This index is
            # also critical for making user lookups fast.
            sa.Index(
                "ix__user__userid",
                _normalise_username(cls.username),
                cls.authority,
                unique=True,
            ),
        )

    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)

    #: Username as chosen by the user on registration
    _username = sa.Column("username", sa.UnicodeText(), nullable=False)

    #: The "authority" for this user. This represents the "namespace" in which
    #: this user lives. By default, all users are created in the namespace
    #: corresponding to `request.domain`, but this can be overridden with the
    #: `h.authority` setting.
    authority = sa.Column("authority", sa.UnicodeText(), nullable=False)

    #: The display name which will be used when rendering an annotation.
    display_name = sa.Column(sa.UnicodeText())

    #: A short user description/bio
    description = sa.Column(sa.UnicodeText())

    #: A free-form column to allow the user to say where they are
    location = sa.Column(sa.UnicodeText())

    #: The user's URI/link on the web
    uri = sa.Column(sa.UnicodeText())

    #: The user's ORCID ID
    orcid = sa.Column(sa.UnicodeText())

    identities = sa.orm.relationship(
        "UserIdentity", backref="user", cascade="all, delete-orphan"
    )

    @hybrid_property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = value

    @username.comparator
    def username(cls):  # noqa: N805
        return UsernameComparator(cls._username)

    @hybrid_property
    def userid(self):
        return "acct:{username}@{authority}".format(
            username=self.username, authority=self.authority
        )

    @userid.comparator
    def userid(cls):  # noqa: N805
        return UserIDComparator(cls.username, cls.authority)

    email = sa.Column(sa.UnicodeText())

    last_login_date = sa.Column(
        sa.TIMESTAMP(timezone=False),
        default=datetime.datetime.utcnow,
        server_default=sa.func.now(),
        nullable=False,
    )
    registered_date = sa.Column(
        sa.TIMESTAMP(timezone=False),
        default=datetime.datetime.utcnow,
        server_default=sa.func.now(),
        nullable=False,
    )

    @sa.orm.validates("email")
    def validate_email(self, key, email):
        if email is None:
            return email

        if len(email) > EMAIL_MAX_LENGTH:
            raise ValueError(
                "email must be less than {max} characters "
                "long".format(max=EMAIL_MAX_LENGTH)
            )
        return email

    @sa.orm.validates("_username")
    def validate_username(self, key, username):
        if not USERNAME_MIN_LENGTH <= len(username) <= USERNAME_MAX_LENGTH:
            raise ValueError(
                "username must be between {min} and {max} "
                "characters long".format(
                    min=USERNAME_MIN_LENGTH, max=USERNAME_MAX_LENGTH
                )
            )

        if not re.match(USERNAME_PATTERN, username):
            raise ValueError(
                "username must have only letters, numbers, " "periods, and underscores."
            )

        return username

    @classmethod
    def get_by_email(cls, session, email, authority):
        """Fetch a user by email address."""
        if email is None:
            return None

        return (
            session.query(cls)
            .filter(
                sa.func.lower(cls.email) == email.lower(), cls.authority == authority
            )
            .first()
        )

    @classmethod
    def get_by_username(cls, session, username, authority):
        """Fetch a user by username."""
        return (
            session.query(cls)
            .filter(cls.username == username, cls.authority == authority)
            .first()
        )

    def __acl__(self):
        terms = []

        # auth_clients that have the same authority as the user
        # may update the user
        user_update_principal = "client_authority:{}".format(self.authority)
        terms.append((security.Allow, user_update_principal, "update"))

        terms.append(security.DENY_ALL)

        return terms

    def __repr__(self):
        return "<User: %s>" % self.username



# -*- coding: utf-8 -*-

"""
Tools for dealing with URIs within the Hypothesis API.

There are two main ways of considering the relationship between annotations and
annotated objects:

1. Annotations are, practically speaking, made on web pages, and thus they have
   a URL associated with them.

2. Annotations are made on documents, and the particular HTML or PDF page being
   annotated is merely a specific manifestation of the abstract document that
   is being annotated. In this scenario, a document may be identified by one or
   more URIs.

The second model is more complex from both a conceptual point of view and in
terms of implementation, but it offers substantial benefits. If we talk of
annotations attached to documents, without regard to presentation format or
location, we are able to do many interesting things:

- Alice makes an annotation on a PDF; Bob visits an HTML version of the same
  document, and sees Alice's annotation.
- Alice makes an annotation on an Instapaper-hosted version of a web page which
  contains a <link rel=canonical> tag. Bob visits the original article and sees
  Alice's annotation.
- Bob makes an annotation on a PDF which is on his local machine. Alice opens
  the same PDF on her machine, and see's Bob's annotations even if the PDF has
  never been uploaded to a webserver. (We can do this because of the
  immutability of PDF documents -- we can uniquely fingerprint each one and
  form a "URN" of the form "urn:x-pdf:<fingerprint>".)

The challenge, then, is to enable these features without making the public API
for creating and updating annotations overly complex. It turns out this is
possible if we can answer two questions:

1. Given two URI strings, do they both refer to the same URI, practically
   speaking? (AKA "normalization".)

   e.g. on the web, the following URLs will *usually* refer to the same web
   page::

       http://example.com/foo?a=hello&b=world
       http://exAMPle.com/foo?a=hello&b=world
       http://example.com/foo/?a=hello&b=world
       http://example.com/foo?b=world&a=hello
       http://example.com/foo?a=hello&b=world#somefragment

2. Given a URI, what are all the known URIs of the underlying *document* (in
   the sense given above). (AKA "expansion".)

   e.g. we may know (from page metadata or otherwise) that all the following
   URIs refer to the same content, even if in differing formats::

       http://example.com/research/papers/2015-discoveries.html
       http://example.com/research/papers/2015-discoveries.pdf
       http://example.org/reprints/example-com-2015-discoveries.pdf
       urn:x-pdf:c83fa94bd1d522276a32f81682a43d29
       urn:doi:10.1000/12345

This package is responsible for defining URI normalization routines for use
elsewhere in the Hypothesis application. URI expansion is handled by
:py:func:`h.storage.expand_uri`.
"""
import re

from hyputils.memex._compat import (
    PY2,
    url_quote,
    url_quote_plus,
    url_unquote,
    url_unquote_plus,
    urlparse,
)

URL_SCHEMES = set(["http", "https"])

# List of regular expressions matching the names of query parameters that we
# strip from URLs as part of normalization.
BLACKLISTED_QUERY_PARAMS = [
    re.compile(regex)
    for regex in set(
        [
            # Google AdWords tracking identifier. Reference:
            #
            #    https://support.google.com/analytics/answer/2938246?hl=en
            #
            r"^gclid$",
            # Google Analytics campaigns. Reference:
            #
            #     https://support.google.com/analytics/answer/1033867?hl=en
            #
            r"^utm_(campaign|content|medium|source|term)$",
            # WebTrends Analytics query params. Reference:
            #
            #     http://help.webtrends.com/en/analytics10/#qpr_about.html
            #
            r"^WT\..+$",
        ]
    )
]

# From RFC3986. The ABNF for path segments is
#
#   path-abempty  = *( "/" segment )
#   ...
#   segment       = *pchar
#   ...
#   pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
#   ...
#   unreserved    = ALPHA / DIGIT / "-" / "." / "_" / "~"
#   sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
#                    / "*" / "+" / "," / ";" / "="
#
# Taken together, this implies the following set of "unreserved" characters for
# path segments (excluding ALPHA and DIGIT which are handled already).
UNRESERVED_PATHSEGMENT = "-._~:@!$&'()*+,;="

# From RFC3986. The ABNF for query strings is
#
#   query         = *( pchar / "/" / "?" )
#
# Where the definition of pchar is as given above.
#
# We exclude "&" and ";" from both names and values, and "=" from names, as
# they are used as delimiters in HTTP URL query strings. In addition, "+" is
# used to denote the space character, so for legacy reasons this is also
# excluded.
UNRESERVED_QUERY_NAME = "-._~:@!$'()*,"
UNRESERVED_QUERY_VALUE = "-._~:@!$'()*,="

# The string that gets prefixed onto a URI if you paste the URI into our Via
# form. For example pasting https://example.com would
# redirect your browser to https://via.hypothes.is/https://example.com.
VIA_PREFIX = "https://via.hypothes.is/"


def normalize(uristr):
    """
    Translate the given URI into a normalized form.

    :type uristr: unicode
    :rtype: unicode
    """

    # In Python 2 functions in urllib expect a byte string whereas in Python 3
    # some functions in urllib work with a byte string or unicode but
    # others (eg. `unquote`) require unicode.
    #
    # Hence we work with byte strings internally in Py 2 and unicode internally
    # in Python 3. In both we always return unicode.
    if PY2:
        uristr = uristr.encode("utf-8")

    def decode_result(result):
        if PY2:
            return result.decode("utf-8")
        else:
            return result

    # Strip proxy prefix for proxied URLs
    for scheme in URL_SCHEMES:
        if uristr.startswith(VIA_PREFIX + scheme + ":"):
            uristr = uristr[len(VIA_PREFIX) :]
            break

    # Try to extract the scheme
    uri = urlparse.urlsplit(uristr)

    # If this isn't a URL, we don't perform any normalization
    if uri.scheme.lower() not in URL_SCHEMES:
        return decode_result(uristr)

    # Don't perform normalization on URLs with no hostname.
    if uri.hostname is None:
        return decode_result(uristr)

    scheme = _normalize_scheme(uri)
    netloc = _normalize_netloc(uri)
    path = _normalize_path(uri)
    query = _normalize_query(uri)
    fragment = None

    uri = urlparse.SplitResult(scheme, netloc, path, query, fragment)

    return decode_result(uri.geturl())


def _normalize_scheme(uri):
    scheme = uri.scheme

    if scheme in URL_SCHEMES:
        scheme = "httpx"

    return scheme


def _normalize_netloc(uri):
    netloc = uri.netloc
    ipv6_hostname = "[" in netloc and "]" in netloc

    username = uri.username
    password = uri.password
    hostname = uri.hostname
    port = uri.port

    # Normalise hostname to lower case
    hostname = hostname.lower()

    # Remove port if default for the scheme
    if uri.scheme == "http" and port == 80:
        port = None
    elif uri.scheme == "https" and port == 443:
        port = None

    # Put it all back together again...
    userinfo = None
    if username is not None:
        userinfo = username
    if password is not None:
        userinfo += ":" + password

    if ipv6_hostname:
        hostname = "[" + hostname + "]"

    hostinfo = hostname
    if port is not None:
        hostinfo += ":" + str(port)

    if userinfo is not None:
        netloc = "@".join([userinfo, hostinfo])
    else:
        netloc = hostinfo

    return netloc


def _normalize_path(uri):
    path = uri.path

    while path.endswith("/"):
        path = path[:-1]

    segments = path.split("/")
    segments = [_normalize_pathsegment(s) for s in segments]
    path = "/".join(segments)

    return path


def _normalize_pathsegment(segment):
    return url_quote(url_unquote(segment), safe=UNRESERVED_PATHSEGMENT)


def _normalize_query(uri):
    query = uri.query

    try:
        items = urlparse.parse_qsl(query, keep_blank_values=True, strict_parsing=True)
    except ValueError:
        # If we can't parse the query string, we better preserve it as it was.
        return query

    # Python sorts are stable, so preserving relative ordering of items with
    # the same key doesn't require any work from us
    items = sorted(items, key=lambda x: x[0])

    # Remove query params that are blacklisted
    items = [i for i in items if not _blacklisted_query_param(i[0])]

    # Normalise percent-encoding for query items
    query = _normalize_queryitems(items)

    return query


def _normalize_queryitems(items):
    segments = [
        "=".join([_normalize_queryname(i[0]), _normalize_queryvalue(i[1])])
        for i in items
    ]
    return "&".join(segments)


def _normalize_queryname(name):
    return url_quote_plus(url_unquote_plus(name), safe=UNRESERVED_QUERY_NAME)


def _normalize_queryvalue(value):
    return url_quote_plus(url_unquote_plus(value), safe=UNRESERVED_QUERY_VALUE)


def _blacklisted_query_param(s):
    """Return True if the given string matches any BLACKLISTED_QUERY_PARAMS."""
    return any(re.match(patt, s) for patt in BLACKLISTED_QUERY_PARAMS)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Generate random strings for use as identifiers in URLs (and other places).

The random strings are generated using an alphabet of ASCII digits and letters
with the following removed:

  - 0, O, I, 1: easily mistaken for one another when reading a URL.
  - C, c, F, f, H, h, S, s, U, u, T, t: avoid making curse words accidentally.

With a default length of n characters, the size of the space of possible
strings is:

    N = len(ALPHABET) ** n

The chance of a collision when generating k random ids of length n is
approximately:

   k**2
   ----
    2N

(for sufficiently large k and N). This means that with a default length of 8,
as here, we have a 99.97% chance of generating 100,000 strings in a row without
collisions, so we hopefully won't need to worry about that any time soon.

Reference: http://preshing.com/20110504/hash-collision-probabilities/
"""

import random

ALPHABET = "123456789ABDEGJKLMNPQRVWXYZabdegijkmnopqrvwxyz"
DEFAULT_LENGTH = 8


def generate(length=DEFAULT_LENGTH):
    """
    Generate a random string of the specified length.

    The returned string is composed of an alphabet that shouldn't include any
    characters that are easily mistakeable for one another (I, 1, O, 0), and
    hopefully won't accidentally contain any English-language curse words.
    """
    return "".join(random.SystemRandom().choice(ALPHABET) for _ in range(length))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from datetime import datetime
import logging

import sqlalchemy as sa
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.ext.hybrid import hybrid_property

from hyputils.memex._compat import urlparse
from hyputils.memex.db import Base, mixins
from hyputils.memex.models.annotation import Annotation
from hyputils.memex.util.uri import normalize as uri_normalize

log = logging.getLogger(__name__)


class ConcurrentUpdateError(Exception):
    """Raised when concurrent updates to document data conflict."""


class Document(Base, mixins.Timestamps):
    __tablename__ = "document"

    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)

    #: The denormalized value of the first DocumentMeta record with type title.
    title = sa.Column("title", sa.UnicodeText())

    #: The denormalized value of the "best" http(s) DocumentURI for this Document.
    web_uri = sa.Column("web_uri", sa.UnicodeText())

    # FIXME: This relationship should be named `uris` again after the
    #        dependency on the annotator-store is removed, as it clashes with
    #        making the Postgres and Elasticsearch interface of a Document
    #        object behave the same way.
    document_uris = sa.orm.relationship(
        "DocumentURI", backref="document", order_by="DocumentURI.updated.desc()"
    )
    meta = sa.orm.relationship(
        "DocumentMeta", backref="document", order_by="DocumentMeta.updated.desc()"
    )

    def __repr__(self):
        return "<Document %s>" % self.id

    def update_web_uri(self):
        """
        Update the value of the self.web_uri field.

        Set self.web_uri to the "best" http(s) URL from self.document_uris.

        Set self.web_uri to None if there's no http(s) DocumentURIs.

        """

        def first_http_url(type_=None):
            """
            Return this document's first http(s) URL of the given type.

            Return None if this document doesn't have any http(s) URLs of the
            given type.

            If no type is given just return this document's first http(s)
            URL, or None.

            """
            for document_uri in self.document_uris:
                uri = document_uri.uri
                if type_ is not None and document_uri.type != type_:
                    continue
                if urlparse.urlparse(uri).scheme not in ["http", "https"]:
                    continue
                return document_uri.uri

        self.web_uri = (
            first_http_url(type_="self-claim")
            or first_http_url(type_="rel-canonical")
            or first_http_url()
        )

    @classmethod
    def find_by_uris(cls, session, uris):
        """Find documents by a list of uris."""
        query_uris = [uri_normalize(u) for u in uris]

        matching_claims = (
            session.query(DocumentURI)
            .filter(DocumentURI.uri_normalized.in_(query_uris))
            .distinct(DocumentURI.document_id)
            .subquery()
        )

        return session.query(Document).join(matching_claims)

    @classmethod
    def find_or_create_by_uris(
        cls, session, claimant_uri, uris, created=None, updated=None
    ):
        """
        Find or create documents from a claimant uri and a list of uris.

        It tries to find a document based on the claimant and the set of uris.
        If none can be found it will return a new document with the claimant
        uri as its only document uri as a self-claim. It is the callers
        responsibility to create any other document uris.
        """

        finduris = [claimant_uri] + uris
        documents = cls.find_by_uris(session, finduris)

        if documents.count() == 0:
            doc = Document(created=created, updated=updated)
            DocumentURI(
                document=doc,
                claimant=claimant_uri,
                uri=claimant_uri,
                type="self-claim",
                created=created,
                updated=updated,
            )
            session.add(doc)

        try:
            session.flush()
        except sa.exc.IntegrityError:
            raise ConcurrentUpdateError("concurrent document creation")

        return documents


class DocumentURI(Base, mixins.Timestamps):
    __tablename__ = "document_uri"
    __table_args__ = (
        sa.UniqueConstraint(
            "claimant_normalized", "uri_normalized", "type", "content_type"
        ),
        sa.Index("ix__document_uri_document_id", "document_id"),
        sa.Index("ix__document_uri_updated", "updated"),
    )

    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)

    _claimant = sa.Column("claimant", sa.UnicodeText, nullable=False)
    _claimant_normalized = sa.Column(
        "claimant_normalized", sa.UnicodeText, nullable=False
    )

    _uri = sa.Column("uri", sa.UnicodeText, nullable=False)
    _uri_normalized = sa.Column(
        "uri_normalized", sa.UnicodeText, nullable=False, index=True
    )

    type = sa.Column(sa.UnicodeText, nullable=False, default="", server_default="")
    content_type = sa.Column(
        sa.UnicodeText, nullable=False, default="", server_default=""
    )

    document_id = sa.Column(sa.Integer, sa.ForeignKey("document.id"), nullable=False)

    @hybrid_property
    def claimant(self):
        return self._claimant

    @claimant.setter
    def claimant(self, value):
        self._claimant = value
        self._claimant_normalized = uri_normalize(value)

    @hybrid_property
    def claimant_normalized(self):
        return self._claimant_normalized

    @hybrid_property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, value):
        self._uri = value
        self._uri_normalized = uri_normalize(value)

    @hybrid_property
    def uri_normalized(self):
        return self._uri_normalized

    def __repr__(self):
        return "<DocumentURI %s>" % self.id


class DocumentMeta(Base, mixins.Timestamps):
    __tablename__ = "document_meta"
    __table_args__ = (
        sa.UniqueConstraint("claimant_normalized", "type"),
        sa.Index("ix__document_meta_document_id", "document_id"),
        sa.Index("ix__document_meta_updated", "updated"),
    )

    id = sa.Column(sa.Integer, autoincrement=True, primary_key=True)

    _claimant = sa.Column("claimant", sa.UnicodeText, nullable=False)
    _claimant_normalized = sa.Column(
        "claimant_normalized", sa.UnicodeText, nullable=False
    )

    type = sa.Column(sa.UnicodeText, nullable=False)
    value = sa.Column(pg.ARRAY(sa.UnicodeText, zero_indexes=True), nullable=False)

    document_id = sa.Column(sa.Integer, sa.ForeignKey("document.id"), nullable=False)

    @hybrid_property
    def claimant(self):
        return self._claimant

    @claimant.setter
    def claimant(self, value):
        self._claimant = value
        self._claimant_normalized = uri_normalize(value)

    @hybrid_property
    def claimant_normalized(self):
        return self._claimant_normalized

    def __repr__(self):
        return "<DocumentMeta %s>" % self.id


def create_or_update_document_uri(
    session, claimant, uri, type, content_type, document, created, updated
):
    """
    Create or update a DocumentURI with the given parameters.

    If an equivalent DocumentURI already exists in the database then its
    updated time will be updated.

    If no equivalent DocumentURI exists in the database then a new one will be
    created and added to the database.

    To be considered "equivalent" an existing DocumentURI must have the same
    claimant, uri, type and content_type, but the Document object that it
    belongs to may be different. The claimant and uri are normalized before
    comparing.

    :param session: the database session
    :type session: sqlalchemy.orm.session.Session

    :param claimant: the .claimant property of the DocumentURI
    :type claimant: unicode

    :param uri: the .uri property of the DocumentURI
    :type uri: unicode

    :param type: the .type property of the DocumentURI
    :type type: unicode

    :param content_type: the .content_type property of the DocumentURI
    :type content_type: unicode

    :param document: the Document that the new DocumentURI will belong to, if a
        new DocumentURI is created
    :type document: h.models.Document

    :param created: the time that will be used as the .created time for the new
        DocumentURI, if a new one is created
    :type created: datetime.datetime

    :param updated: the time that will be set as the .updated time for the new
        or existing DocumentURI
    :type updated: datetime.datetime

    """
    docuri = (
        session.query(DocumentURI)
        .filter(
            DocumentURI.claimant_normalized == uri_normalize(claimant),
            DocumentURI.uri_normalized == uri_normalize(uri),
            DocumentURI.type == type,
            DocumentURI.content_type == content_type,
        )
        .first()
    )

    if docuri is None:
        docuri = DocumentURI(
            claimant=claimant,
            uri=uri,
            type=type,
            content_type=content_type,
            document=document,
            created=created,
            updated=updated,
        )
        session.add(docuri)
    elif not docuri.document == document:
        log.warning(
            "Found DocumentURI (id: %d)'s document_id (%d) doesn't match "
            "given Document's id (%d)",
            docuri.id,
            docuri.document_id,
            document.id,
        )

    docuri.updated = updated

    try:
        session.flush()
    except sa.exc.IntegrityError:
        raise ConcurrentUpdateError("concurrent document uri updates")


def create_or_update_document_meta(
    session, claimant, type, value, document, created, updated
):
    """
    Create or update a DocumentMeta with the given parameters.

    If an equivalent DocumentMeta already exists in the database then its value
    and updated time will be updated.

    If no equivalent DocumentMeta exists in the database then a new one will be
    created and added to the database.

    To be considered "equivalent" an existing DocumentMeta must have the given
    claimant and type, but its value, document and created and updated times
    needn't match the given ones.

    :param session: the database session
    :type session: sqlalchemy.orm.session.Session

    :param claimant: the value to use for the DocumentMeta's claimant attribute
        if a new DocumentMeta is created
    :type claimant: unicode

    :param type: the value of the new or existing DocumentMeta's type attribute
    :type type: unicode

    :param value: the value to set the new or existing DocumentMeta's value
        attribute to
    :type value: list of unicode strings

    :param document: the value to use for the DocumentMeta's document if a new
        DocumentMeta is created
    :type document: h.models.Document

    :param created: the value to use for the DocumentMeta's created attribute
        if a new DocumentMeta is created
    :type created: datetime.datetime

    :param updated: the value to set the new or existing DocumentMeta's updated
        attribute to
    :type updated: datetime.datetime

    """
    existing_dm = (
        session.query(DocumentMeta)
        .filter(
            DocumentMeta.claimant_normalized == uri_normalize(claimant),
            DocumentMeta.type == type,
        )
        .one_or_none()
    )

    if existing_dm is None:
        session.add(
            DocumentMeta(
                claimant=claimant,
                type=type,
                value=value,
                document=document,
                created=created,
                updated=updated,
            )
        )
    else:
        existing_dm.value = value
        existing_dm.updated = updated
        if not existing_dm.document == document:
            log.warning(
                "Found DocumentMeta (id: %d)'s document_id (%d) doesn't "
                "match given Document's id (%d)",
                existing_dm.id,
                existing_dm.document_id,
                document.id,
            )

    if type == "title" and value and not document.title:
        document.title = value[0]

    try:
        session.flush()
    except sa.exc.IntegrityError:
        raise ConcurrentUpdateError("concurrent document meta updates")


def merge_documents(session, documents, updated=None):
    """
    Takes a list of documents and merges them together. It returns the new
    master document.

    The support for setting a specific value for the `updated` should only
    be used during the Postgres migration. It should be removed afterwards.
    """
    if updated is None:
        updated = datetime.utcnow()

    master = documents[0]
    duplicates = documents[1:]
    duplicate_ids = [doc.id for doc in duplicates]

    for doc in duplicates:
        for _ in range(len(doc.document_uris)):
            u = doc.document_uris.pop()
            u.document = master
            u.updated = updated

        for _ in range(len(doc.meta)):
            m = doc.meta.pop()
            m.document = master
            m.updated = updated

    try:
        session.flush()
        session.query(Annotation).filter(
            Annotation.document_id.in_(duplicate_ids)
        ).update({Annotation.document_id: master.id}, synchronize_session="fetch")
        session.query(Document).filter(Document.id.in_(duplicate_ids)).delete(
            synchronize_session="fetch"
        )
    except sa.exc.IntegrityError:
        raise ConcurrentUpdateError("concurrent document merges")

    return master


def update_document_metadata(
    session,
    target_uri,
    document_meta_dicts,
    document_uri_dicts,
    created=None,
    updated=None,
):
    """
    Create and update document metadata from the given annotation.

    Document, DocumentURI and DocumentMeta objects will be created, updated
    and deleted in the database as required by the given annotation and
    document meta and uri dicts.

    :param target_uri: the target_uri of the annotation from which the document metadata comes from
    :type target_uri: unicode

    :param document_meta_dicts: the document metadata dicts that were derived
        by validation from the "document" dict that the client posted
    :type document_meta_dicts: list of dicts

    :param document_uri_dicts: the document URI dicts that were derived by
        validation from the "document" dict that the client posted
    :type document_uri_dicts: list of dicts

    :param created: Date and time value for the new document records
    :type created: datetime.datetime

    :param updated: Date and time value for the new document records
    :type updated: datetime.datetime

    :returns: the matched or created document
    :rtype: h.models.Document
    """
    if created is None:
        created = datetime.utcnow()
    if updated is None:
        updated = datetime.utcnow()

    documents = Document.find_or_create_by_uris(
        session,
        target_uri,
        [u["uri"] for u in document_uri_dicts],
        created=created,
        updated=updated,
    )

    if documents.count() > 1:
        document = merge_documents(session, documents, updated=updated)
    else:
        document = documents.first()

    document.updated = updated

    for document_uri_dict in document_uri_dicts:
        create_or_update_document_uri(
            session=session,
            document=document,
            created=created,
            updated=updated,
            **document_uri_dict
        )

    document.update_web_uri()

    for document_meta_dict in document_meta_dicts:
        create_or_update_document_meta(
            session=session,
            document=document,
            created=created,
            updated=updated,
            **document_meta_dict
        )

    return document
